unit IEComboBox;
{==================================================================

 Components TIECustomCombobox  /  Version 1.3  / 06.1999
            TIEComboBox
            TIEDriveBox
 ======================================================


    Description:
    ============
    TIECustomComboBox is a combobox with variable width of the dropdown list and
    provides the additional event OnCloseUp.

    TIEComboBox publishes the properties of the class TIECustomComboBox including
    the property Align wich might be was forgotten by Borland.

    TIEDriveComboBox realizes a selection control for the aviable drives of
    the system with icons. The drive icons are taken from the system image
    list.


    Author:
    =======
    (c) Ingo Eckel 1999
    Sodener Weg 38
    65812 Bad Soden
    Germany

    For detailed documentation and history see the documentation in TIEDriveComboBox.htm.

    V1.3:
    - Property DisplayStyle changed.


{==================================================================}

{Required compiler options:}
{$A+,B-,X+,H+,P+}


Interface

Uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, Math, Imglist, ShellAPI,
  IEDriveInfo,
  DriveView;

Const IconWidth = 16;

type
  TIECloseUpEvent = procedure (Sender: TObject; SelectionOK : Boolean) of object;
  TDriveType      = (dtUnknown, dtNoRootDrive, dtFloppy, dtFixed, dtRemote, dtCDROM, dtRAM);
  TDriveTypes     = Set Of TDrivetype;

// =======================================================================
// Class TIECustomComboBox
// =======================================================================
  TIECustomComboBox = class(TCustomComboBox)
  Private
    { Private-Deklarationen }
    fOnCloseUp   : TIECloseUpEvent;
    FItemWidth   : Integer;
    FDropDownFixedWidth: Integer;
    FSelectionOK : Boolean;

    Procedure SetDropDownFixedWidth(const Value: Integer);
    Function GetTopIndex : Integer;
    Procedure SetTopIndex(i : Integer);

  Protected
    { Protected-Deklarationen }
    Procedure CNCommand(Var Message: TWMCommand); Message CN_COMMAND;
    Procedure DropDown;                           Override;
    Function  GetMaxItemWidth : Integer;          Virtual;

  Public
    { Public-Deklarationen }
    Constructor Create(AOwner : TComponent); Override;
    Property ItemWidth : Integer Read FItemWidth Write FItemWidth;
    Property TopIndex : Integer  Read GetTopIndex Write SetTopIndex;

  Published
    { Published-Deklarationen }
    Property OnCloseUp : TIECloseUpEvent  Read fOnCloseUp
                                          Write fOnCloseUp;
    Property DropDownFixedWidth : Integer Read FDropDownFixedWidth
                                          Write SetDropDownFixedWidth;

    Function GetTextWidth(T : String) : Integer;

  End;


// =======================================================================
// Class TIEComboBox
// =======================================================================
  TIEComboBox = class(TIECustomComboBox)

  Published
    { Published-Deklarationen }
    Property Align;
    property Style; {Must be published before Items}
    property Anchors;
    property BiDiMode;
    property Color;
    property Constraints;
    property Ctl3D;
    property DragCursor;
    property DragKind;
    property DragMode;
    property DropDownCount;
    property Enabled;
    property Font;
    property ImeMode;
    property ImeName;
    property ItemHeight;
    property Items;
    property MaxLength;
    property ParentBiDiMode;
    property ParentColor;
    property ParentCtl3D;
    property ParentFont;
    property ParentShowHint;
    property PopupMenu;
    property ShowHint;
    property Sorted;
    property TabOrder;
    property TabStop;
    property Text;
    property Visible;
    property OnChange;
    property OnClick;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnDrawItem;
    property OnDropDown;
    property OnEndDock;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyUp;
    property OnMeasureItem;
    property OnStartDock;
    property OnStartDrag;
  end;

// =======================================================================
// Class TIEDriveComboBox
// =======================================================================
  TIEDriveComboBox = class(TIECustomComboBox)
  Private
    { Private-Deklarationen }
    FDrive           : TDrive;
    fDisplayStyle    : TVolumeDisplayStyle;
    fImgList         : TImageList;
    FDriveView       : TDriveView;
    FDriveTypes      : TDriveTypes;

    Procedure SetDrive(Drive : TDrive);
    Procedure SetDisplayStyle(Options: TVolumeDisplayStyle);
    Procedure ResetItemHeight;                    Virtual;


  Protected
    { Protected-Deklarationen }

    Property Style Default csOwnerDrawFixed;
    Property Items Stored False;

    Property ItemHeight Stored False;
    Property MaxLength Stored False;
    Property Sorted Stored False;
    Property Text Stored False;

    Function  GetDisplayName(Drive : TDrive) : String;                       Virtual;
    Procedure SetDriveView(DriveView : TDriveView);                          Virtual;
    Procedure SetDriveTypes(DriveTypes : TDriveTypes);                       Virtual;
    Function  GetMaxItemWidth : Integer;                                     Override;
    Procedure SetImageList;
    Procedure CMFontChanged(Var Message: TMessage); Message CM_FONTCHANGED;
    Procedure Notification(AComponent: TComponent; Operation: TOperation);   Override;
    Procedure CreateWnd;                                                     Override;
    Procedure DrawItem(Index: Integer; Rect: TRect; State: TOwnerDrawState); Override;
    Procedure Change;                                                        Override;


  Public
    { Public-Deklarationen }
    Constructor Create(AOwner : TComponent); Override;
    Destructor Destroy;                      Override;
    Procedure ResetItems;
    Function  GetDriveIndex(Drive : TDrive) : Integer;

  Published
    { Published-Deklarationen }
    Property Drive : TDrive Read FDrive
                            Write SetDrive
                            Default 'C';
    Property DisplayStyle  : TVolumeDisplayStyle Read fDisplayStyle
                                                 Write SetDisplayStyle
                                                 Default doPrettyName;
    Property DriveView : TDriveView Read FDriveView
                                    Write SetDriveView;
    Property DriveTypes : TDriveTypes Read FDriveTypes
                                      Write SetDriveTypes
                                      Default [dtFloppy, dtFixed, dtRemote, dtCDROM, dtRAM];
    property Align;
    property Anchors;
    property BiDiMode;
    property Color;
    property Constraints;
    property Ctl3D;
    property DragCursor;
    property DragKind;
    property DragMode;
    property DropDownCount;
    property Enabled;
    property Font;
    property ImeMode;
    property ImeName;
    property ParentBiDiMode;
    property ParentColor;
    property ParentCtl3D;
    property ParentFont;
    property ParentShowHint;
    property PopupMenu;
    property ShowHint;
    property TabOrder;
    property TabStop;
    property Visible;
    property OnChange;
    property OnClick;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnDrawItem;
    property OnDropDown;
    property OnEndDock;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyUp;
    property OnStartDock;
    property OnStartDrag;
  End;


Function  GetItemHeight(Font: TFont): Integer;
procedure Register;

implementation

// =======================================================================
// Class TIECustomComboBox
// =======================================================================
constructor TIECustomComboBox.Create(AOwner : TComponent);
begin
  inherited Create(AOwner);
  fSelectionOK := False;
end; {TIECustomComboBox.Create}


Procedure TIECustomComboBox.DropDown;
Begin
  {Check to see if DropDownFixed Width > 0. Then just set the
   width of the list box. Otherwise, loop through the items
   and set the width of the list box to 8 pixels > than the
   widest string to buffer the right side. Anything less than
   8 for some reason touches the end of the item on high-res
   monitor settings.}
  If (FDropDownFixedWidth > 0) Then
    Self.Perform(CB_SETDROPPEDWIDTH, FDropDownFixedWidth, 0)
  Else
    Begin
      ItemWidth := GetMaxItemWidth + 8;
      Self.Perform(CB_SETDROPPEDWIDTH, ItemWidth, 0);
    end;
  Inherited DropDown;
end; {TIECustomComboBox.DropDown}


Function TIECustomComboBox.GetTextWidth(T : String) : Integer;
Var DC: HDC;
    SaveFont: HFont;
    Size    : TSize;

Begin
  DC := GetDC(0);
  Try
    SaveFont := SelectObject(DC, Font.Handle);
    GetTextExtentPoint32(DC, PChar(T), Length(T), Size);
    Result := Size.Cx;
    SelectObject(DC, SaveFont);
  Finally
    ReleaseDC(0, DC);
  End;
end; {TIECustomComboBox.GetTextWidth}


Function TIECustomComboBox.GetMaxItemWidth : Integer;
var DC: HDC;
    SaveFont: HFont;
    Size    : TSize;
    i       : Integer;

Begin
  Result := 0;
  DC := GetDC(0);
  Try
    SaveFont := SelectObject(DC, Font.Handle);
    For i := 0 To Items.Count - 1 Do
    Begin
      GetTextExtentPoint32(DC, PChar(Items[i]), Length(Items[i]), Size);
      If Size.Cx > Result Then
        Result := Size.Cx;
    End;
    SelectObject(DC, SaveFont);
  Finally
    ReleaseDC(0, DC);
  End;
end; {TIECustomComboBox.GetMaxItemWidth}


Procedure TIECustomComboBox.SetDropDownFixedWidth(const Value: Integer);
begin
  FDropDownFixedWidth := Value;
end; {TIECustomComboBox.SetDropDownFixedWidth}


Function TIECustomComboBox.GetTopIndex : Integer;
Begin
  Result := Perform(CB_GETTOPINDEX, 0, 0);
End; {TIECustomComboBox.GetTopIndex}

Procedure TIECustomComboBox.SetTopIndex(i : Integer);
Begin
  Perform(CB_SETTOPINDEX, i, 0);
End; {TIECustomComboBox.SetTopIndex}


Procedure TIECustomComboBox.CNCommand(Var Message: TWMCommand);
Begin
  Inherited;
  Case Message.NotifyCode Of
    CBN_CLOSEUP:
    Begin
      IF Assigned(fOnCloseUp) Then
      fOnCloseUp(Self, fSelectionOK);
    End;
    CBN_SELENDCANCEL:
      fSelectionOK := False;
    CBN_SELENDOK:
      fSelectionOK := True;
  End;
End; {TIECustomComboBox.CNCommand}



// =======================================================================
// Class TIEDriveComboBox
// =======================================================================
Constructor TIEDriveComboBox.Create(AOwner : TComponent);
Begin
  Inherited Create(AOwner);
  Style := csOwnerDrawFixed;
  FDrive := FirstFixedDrive;
  FDriveTypes := [dtFloppy, dtFixed, dtRemote, dtCDROM, dtRAM];
End; {TIEDriveComboBox.Create}


Destructor TIEDriveComboBox.Destroy;
Begin
  fImgList.Free;
  Inherited Destroy;
End; {TIEDriveComboBox.Destroy}


Procedure TIEDriveComboBox.CreateWnd;
Begin
  Inherited CreateWnd;
  SetImageList;
  ResetItemHeight;
  ResetItems;
End; {TIEDriveComboBox.CreateWnd}


Procedure TIEDriveComboBox.CMFontChanged(Var Message: TMessage);
Begin
  Inherited;
  ResetItemHeight;
  RecreateWnd;
End; {TIEDriveComboBox.CMFontChanged}


Procedure TIEDriveComboBox.Change;
begin
  fDrive := Items[ItemIndex][1];
  IF Assigned(FDriveView) And
    (Not Assigned(FDriveView.Selected) Or (fDrive <> FDriveView.GetDrive(FDriveView.Selected))) Then
    With FDriveView Do
  Begin
    Directory := Upcase(FDrive)  + ':\';
    IF Assigned(Selected) Then
      Selected.Expand(False);
  End;
  Inherited Change;
end; {Change}


Function TIEDriveComboBox.GetDisplayName(Drive : TDrive) : String;
Begin
  Case DisplayStyle Of
    doPrettyName:      Result := DriveInfo.GetPrettyName(Drive);
    doDisplayName:     Result := DriveInfo.GetDisplayName(Drive);
    doLongPrettyName:  Result := DriveInfo.GetLongPrettyName(Drive);
  End;
End; {TIEDriveComboBox.GetDisplayName}


Function TIEDriveComboBox.GetDriveIndex(Drive : TDrive) : Integer;
Var i : Integer;
Begin
  Result := -1;
  For i := 0 to Pred(Items.Count) Do
  Begin
    IF Items[i][1] = Drive Then
    Begin
      Result := i;
      Break;
    End;
  End;
End; {TIEDriveComboBox.GetDriveIndex}


Procedure TIEDriveComboBox.ResetItems;
Var D : TDrive;

Begin
  Items.BeginUpdate;
  Items.Clear;
  For D:= FirstDrive To LastDrive Do
  Begin
    IF DriveInfo[D].Valid And (TDriveType(DriveInfo[D].DriveType) in FDriveTypes) Then
    Items.Add(D + GetDisplayName(D));
  End;
  ItemIndex := GetDriveIndex(fDrive);
  IF (ItemIndex < 0) And (Items.Count > 0) Then
  ItemIndex := 0;
  Items.EndUpdate;
End; {TIEDriveComboBox.ResetItems }


Procedure TIEDriveComboBox.SetDrive(Drive : TDrive);
Var NewIndex : Integer;
Begin
  Drive := Upcase(Drive);
  IF Drive <> FDrive Then
  Begin
    DriveInfo.ReadDriveStatus(Drive, dsValid);
    IF DriveInfo[Drive].Valid And (TDriveType(DriveInfo[Drive].DriveType) in FDriveTypes) Then
    Begin
      FDrive := Drive;
      NewIndex := GetDriveIndex(Drive);
      IF NewIndex >= 0 Then
      ItemIndex := NewIndex;
    End;
  End;
End; {TIEDriveComboBox.SetDrive}


Procedure TIEDriveComboBox.SetDriveView(DriveView : TDriveView);
Begin
  IF Assigned(FDriveView) Then
  FDriveView.DriveBox := NIL;
  FDriveView := DriveView;
  IF Assigned(FDriveView) Then
  FDriveView.DriveBox := Self;
End; {TIEDriveComboBox.SetDriveView}


Procedure TIEDriveComboBox.SetDriveTypes(DriveTypes : TDriveTypes);
Begin
  IF FDriveTypes <> DriveTypes Then
  Begin
    FDriveTypes := DriveTypes;
    ResetItems;
  End;
End; {TIEDriveComboBox.SetDriveTypes}


Procedure TIEDriveComboBox.Notification(AComponent: TComponent; Operation: TOperation);
begin
  Inherited Notification(AComponent, Operation);
  If (Operation = opRemove) and (AComponent = FDriveView) then
  FDriveView := NIL;
end; {Notification}


Procedure TIEDriveComboBox.SetDisplayStyle(Options: TVolumeDisplayStyle);
Begin
  IF fDisplayStyle <> Options Then
  Begin
    fDisplayStyle := Options;
    ResetItems;
    Invalidate;
  End;
End; {TIEDriveComboBox.SetDisplayStyle}


Procedure TIEDriveComboBox.SetImageList;
Var SysIL: UInt;
    SFI:   TSHFileInfo;

Begin
  fimgList := TImageList.Create(self);
  SysIL := SHGetFileInfo('', 0, SFI, SizeOf(SFI), SHGFI_SYSICONINDEX or SHGFI_SMALLICON);
  if SysIL <> 0 then
  begin
    fimgList.ShareImages  := True;
    fimgList.Handle := SysIL;
    fImgList.drawingstyle := dsTransparent;
  End;
End; {TIEDriveComboBox.SetImageList}


procedure TIEDriveComboBox.DrawItem(Index: Integer; Rect: TRect; State: TOwnerDrawState);
Var Text: array[0..255] of Char;
    Drive : TDrive;

Begin
  With Canvas Do
  Begin
    FillRect(Rect);
    Pen.Color := clWindowText;
    Drive := Items[Index][1];

    StrPCopy(Text, GetDisplayName(Drive));

    FImgList.Draw(Canvas, Rect.Left + 2, Rect.Top , DriveInfo.GetImageIndex(Drive));

    Rect.Left := Rect.Left + IconWidth + 6;
    DrawText(Canvas.Handle, Text, StrLen(Text), Rect,
    DrawTextBiDiModeFlags(DT_SINGLELINE or DT_VCENTER or DT_NOPREFIX));
  end;
end; {TIEDriveComboBox.DrawItem}


Procedure TIEDriveComboBox.ResetItemHeight;
Var H: Integer;

Begin
  H := Max(GetItemHeight(Font), 10) + 2;
  Inherited ItemHeight := H;
  If HandleAllocated Then
  Begin
    {Set height of list items:}
    SendMessage(Handle, CB_SETITEMHEIGHT,  0, Max(H, 12));
    {Set height of selection field:}
    SendMessage(Handle, CB_SETITEMHEIGHT, -1, H);
    {Set height of delphi-control:}
    Height := H;
  End;
end; {TIEDriveComboBox.ResetItemHeight}


Function TIEDriveComboBox.GetMaxItemWidth : Integer;
Begin
  Result := Inherited GetMaxItemWidth + IconWidth + 6;
End; {TIEDriveComboBox.GetMaxItemWidth}


// =======================================================================

Function GetItemHeight(Font: TFont): Integer;
var DC: HDC;
    SaveFont: HFont;
    Metrics: TTextMetric;
Begin
  DC := GetDC(0);
  Try
    SaveFont := SelectObject(DC, Font.Handle);
    GetTextMetrics(DC, Metrics);
    SelectObject(DC, SaveFont);
  Finally
    ReleaseDC(0, DC);
  End;
  Result := Metrics.tmHeight;
End; {GetItemHeight}


procedure Register;
begin
  RegisterComponents('IE', [TIEComboBox, TIEDriveComboBox]);
end;

end.
